/*
 * This module implements a directory /proc/foobar in the procfs and
 * two files in the directory to report some not really interesting
 * data (such as how long ago a file was read).
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/jiffies.h>

MODULE_AUTHOR("Juergen Schoenwaelder");
MODULE_LICENSE("Dual BSD/GPL");
MODULE_DESCRIPTION("Simple procfs demonstration module.");

#define kwarn(FMT, ...) \
        printk(KERN_WARNING "[%s] "FMT"\n", __this_module.name, ##__VA_ARGS__)

static const char *dir_name = "foobar";
static struct proc_dir_entry *dir_entry;

struct file_data {
    const char *name;
    unsigned long lasttime;
};

static struct file_data files[] = {
    { .name = "foo", .lasttime = 0 },
    { .name = "bar", .lasttime = 0 },
    { .name = NULL,  .lasttime = 0 },
};

static int
file_show(struct seq_file *m, void *v)
{
    struct file_data *data = m->private;
    unsigned long now = jiffies;

    if (! data) {
        kwarn("show for a file without data");
        return 0;
    }

    seq_printf(m, "%s", data->name);
    if (data->lasttime) {
        seq_printf(m, " (last accessed %lu seconds ago)",
		   (now - data->lasttime) / HZ);
    }
    data->lasttime = now;
    seq_printf(m, "\n");
      
    return 0;
}

static int
file_open(struct inode *inode, struct file *file)
{
    int i;
    struct file_data *data = NULL;

    if (!file 
	|| !file->f_path.dentry
	|| !file->f_path.dentry->d_name.name) {
        kwarn("open for a file without a name");
        return -ENOENT;
    }

    for (i = 0; files[i].name; i++) {
        if (strcmp(files[i].name, file->f_path.dentry->d_name.name) == 0) {
	    data = &files[i];
	    break;
	}
    }

    /* The last argument passed to single_open() will be available
     * in the private member of the seq_file structure. */
    return single_open(file, file_show, data);
}

static const struct file_operations procfs_file_fops = {
    .owner      = THIS_MODULE,
    .open       = file_open,
    .read       = seq_read,
    .llseek     = seq_lseek,
    .release    = single_release,
};

static int __init
procfs_init(void)
{
    struct proc_dir_entry *entry;
    int i;
    
    dir_entry = proc_mkdir(dir_name, NULL);
    if (! dir_entry) {
        kwarn("unable to create directory %s", dir_name);
        return -ENOMEM;
    }

    for (i = 0; files[i].name; i++) {
        entry = create_proc_entry(files[i].name, 0, dir_entry);
	if (! entry) {
	    kwarn("unable to create file %s - skipping", files[i].name);
	    continue;
	}
	entry->proc_fops = &procfs_file_fops;
    }

    return 0;
}

static void __exit
procfs_exit(void)
{
    int i;

    for (i = 0; files[i].name; i++) {
        remove_proc_entry(files[i].name, dir_entry);
    }

    remove_proc_entry(dir_name, NULL);
}

module_init(procfs_init);
module_exit(procfs_exit);
